﻿import asyncio
from datetime import datetime as dt
from colorama import Fore

from rasa.core.agent         import Agent 
from rasa.core.interpreter   import RasaNLUInterpreter 
from rasa.utils.endpoints    import EndpointConfig
from rasa.shared.core.events import (BotUttered, UserUttered,ActionExecuted, SessionStarted, DefinePrevUserUtteredFeaturization, SlotSet)

#-----------------------------------------------------------
# 
def show_slots(tracker, all = False):
    """ Вывести список слотов и их значений """
    print('slots:')
    if  tracker.slots:
        for s in tracker.slots:
            if s != 'session_started_metadata' or all:
                print(Fore.YELLOW, ("- %-12s" % (s))+Fore.WHITE+f': {tracker.slots[s]}')    

#-----------------------------------------------------------
#                 
def show_entity(en):
    """ Вернуть все поля сущности как строку """
    res = ""
    if 'entity'            in en: res += Fore.YELLOW+"   - %-12s  " % (en['entity']) + Fore.WHITE + " "
    if 'confidence_entity' in en: res += "[%.2f]" % (en['confidence_entity'])+Fore.WHITE + " "
    if 'start' in en and 'end' in en: res += "pos:[%2d,%2d] " % (en['start'],en['end'])
    if 'value'             in en: res += f"value: {Fore.YELLOW}{en['value']}"+Fore.WHITE + " "
    if 'role'              in en: res += f"role: {Fore.YELLOW}{en['role']}"+Fore.WHITE + " "
    if 'group'             in en: res += f"group: {Fore.YELLOW}{en['group']}"+Fore.WHITE + " "
    if 'extractor'         in en: res == en['extractor'] + " "
    
    return res  
#-----------------------------------------------------------
# 
def print_event(events, is_text=True, is_slotset=True):          
    """ вывести краткую информацию о всех событиях events """
    res = []
    for ev in events:
        if type(ev) is SessionStarted:
            continue
        if type(ev) is DefinePrevUserUtteredFeaturization:
            continue
        if type(ev) is ActionExecuted:
            if ev.action_name == 'action_listen' or ev.action_name == 'action_session_start': 
                continue
            res.append(Fore.WHITE+"- action: "+"%-32s  < " % (ev.action_name)  +  "[%.2f] %s" % (ev.confidence, ev.policy ))        
        elif    type(ev) is BotUttered:
            if is_text:
                res.append("    "+Fore.GREEN+ ev.text+ Fore.WHITE)
        elif  type(ev) is UserUttered:
            res.append(Fore.WHITE +"- intent: " +  ev.intent['name'] + " [%.2f]" %( ev.intent['confidence'])+ Fore.CYAN + "\n    "+ ev.text+ Fore.WHITE)
            if ev.entities:
                for e in ev.entities:
                    res.append("     "+show_entity(e))
        else:
            if type(ev) is SlotSet and not is_slotset:
                continue
            res.append("- action: " + str(ev))   
    return res

#-----------------------------------------------------------
# 
def show_events(tracker, all = False):                
    """ Вывести все события """    
    if tracker and tracker.events:
        if all:
            print("tracker.events: ", tracker.events)
            print("----------------------------------------")
            for ev in tracker.events:            
                print(" - ", ev)
        else:
            print('events:')                    
            res = print_event(tracker.events)
            for r in res:
                print(r)
#-----------------------------------------------------------
#
def show_short_info(tracker, num):
    """ Вывести краткую информацию """
    if tracker and tracker.latest_message:
        print("----------------------------------------")
        res = ""
        mes = tracker.latest_message
        if mes.intent:
            print(Fore.WHITE+f'- intent:', end="")           
            if 'name'       in mes.intent and 'confidence' in mes.intent:
                print(Fore.CYAN, "%-32s  < " % (mes.intent['name'])+ Fore.WHITE+ "[%.2f]" % (mes.intent['confidence'])); 
                res = "%s[%.3f]" % (mes.intent['name'], mes.intent['confidence'])                             
        
        if mes.entities and len(mes.entities) > 0:
            print( '  entities:')   
            for en in mes.entities:
                print(show_entity(en))
        if tracker.events and len(tracker.events) > 3:              
            lst = print_event(tracker.events, False, False)
            for i in range(max(0,len(lst)-num), len(lst)):
                print(lst[i])
        print("----------------------------------------")
        return res
            
#-----------------------------------------------------------
#
def show_info(agent, human, SHOW_DEBUG_INFO):
    """ Отладочная информация бота """
    tracker = agent.tracker_store.retrieve('default')
    if not tracker or not tracker.latest_message:
        return
        
    if len(human) > 1 and 'iш'.find(human[1]) >= 0:
        SHOW_DEBUG_INFO = not SHOW_DEBUG_INFO
        if SHOW_DEBUG_INFO:
            show_short_info(tracker)
        return SHOW_DEBUG_INFO
        
    print(Fore.WHITE, end="")
    print("----------------------------------------")        
    if  len(human) > 1  and 'sы'.find(human[1]) >= 0:
        show_slots(tracker)
    elif len(human) > 1 and 'eу'.find(human[1]) >= 0:
        show_events(tracker)        
    elif len(human) > 1 and 'aф'.find(human[1]) >= 0:
        print(f'sender_id:{tracker.sender_id}')
        print(f'text:     {tracker.latest_message.text}')   
        show_short_info(tracker, 10000)
        print(f'channel:  {tracker.latest_message.input_channel}')   
        print(f'action:   {tracker.latest_action_name}')
        print(f'loops:    {tracker.active_loop}')        
        show_slots(tracker , all=True)
        show_events(tracker, all=True)                    
    else:
        print(f'Unknown command: {human}')
    return SHOW_DEBUG_INFO

#-----------------------------------------------------------
#
def show_help():
    """ Вывести информацию о управляющих командах """
    print(Fore.YELLOW)
    print('You can enter the following commands at any time:\n')
    print('== for exit (double equality)') 
    print('=i on/off short debug information')
    print('=s get list of slots')
    print('=e get list of events')    
    print('=a get all of debug information')
    print('or just start chatting with the bot :)')
    print("----------------------------------------\n")
#-----------------------------------------------------------
#
def load_model(path_str="models"):
    """ Загрузка модели """
    from rasa.cli.utils        import get_validated_path    
    from rasa.model            import get_model
    from rasa.shared.constants import DEFAULT_MODELS_PATH 

    print(Fore.YELLOW, "Loading model... ",  end="")
    
    action_endpoint = EndpointConfig(url="http://localhost:5055/webhook")    
    model = get_validated_path(path_str, "model", DEFAULT_MODELS_PATH)
    model_path = get_model(model)       
        
    print("model_path: ", model_path)
    print(" Start session (wait a bit)", Fore.WHITE)
    
    return Agent.load(model_path, action_endpoint = action_endpoint) 

#-----------------------------------------------------------
#
async def main():
    """ Основная петля взаимодействия с ботом """
    SHOW_DEBUG_INFO = True    # выводить отладочную информацию после каждого ответ
    
    agent = load_model()                    # загружаем RASA модель 
    show_help()                                 
    
    fout = open("!agent.log", 'a+', encoding='utf-8')      
    fout.write(dt.now().strftime("%Y-%m-%d %H:%M:%S")+"\n\n")
    
    while True:
        print(Fore.CYAN, "-> ", end="")
        human = input()                     # ввод человека
        
        if len(human) == 0:                 # пустая строка
            continue
        if '=' == human[0]:                 # служебные команды
            if human[0:2] == '==':
                break
            SHOW_DEBUG_INFO = show_info(agent, human, SHOW_DEBUG_INFO)
            continue        

        lst = await agent.handle_text(human)# ответ бота            

        inf = ""
        if SHOW_DEBUG_INFO:
            inf = show_short_info(agent.tracker_store.retrieve('default'), len(lst))
        
        fout.write("-> "+human+"    | "+inf+"\n")            
        
        for t in lst:            
            print(Fore.GREEN +"<= ",t['text'],Fore.WHITE)       
            fout.write("<= "+t['text']+"\n")            
                                                   
    fout.write("--------------------------------------------------\n")
    fout.close()
    
asyncio.run( main() )