from rasa.cli.utils                import get_validated_path
from rasa.model                    import get_model, get_model_subdirectories
from rasa.core.interpreter         import RasaNLUInterpreter

from rasa.nlu.model                import Interpreter
from rasa.shared.nlu.interpreter   import RegexInterpreter, INTENT_MESSAGE_PREFIX
from rasa.shared.utils.cli         import print_success

from rasa.shared.nlu.constants                import TEXT, INTENT
from rasa.shared.nlu.training_data.message    import Message
from rasa.nlu.tokenizers.whitespace_tokenizer import WhitespaceTokenizer

import rasa.nlu.run
import pathlib

from rasa.shared.constants import (
    DOCS_BASE_URL,
    DEFAULT_ENDPOINTS_PATH,
    DEFAULT_CREDENTIALS_PATH,
    DEFAULT_ACTIONS_PATH,
    DEFAULT_MODELS_PATH,
)
#######################################################################

def load_interpreter(model_dir="", model="model"):
    """Загружаем NLU-модель"""
    path_str = str(pathlib.Path(model_dir) / model)
    model = get_validated_path(path_str, "model", DEFAULT_MODELS_PATH)
    model_path = get_model(model)
    _, nlu_model = get_model_subdirectories(model_path)
    
    print_success("DEFAULT_MODELS_PATH:"+DEFAULT_MODELS_PATH)
    print_success("model:"+model)
    print_success("model_path:"+model_path)
        
    return RasaNLUInterpreter(nlu_model)

def get_tokens(attribute=TEXT):
    """Токены, получаемые WhitespaceTokenizer()"""
    tokenizer = WhitespaceTokenizer()    
    tokens = tokenizer.tokenize(Message.build(text=message), attribute=attribute) 
    print("tokens:|", end='')
    for t in tokens:
        print(t.text, end="|")    
        if len(t.data) > 0:            
            print(t.data, end="]")        
    print("\n")

def get_pipeline():
    """Вывести информацию о конвейере"""
    fout = open("!nlu_debug.out", 'w', encoding='utf-8')      
    msg = Message({TEXT: message})
    was_takens = False
    for i,p in enumerate(interpreter.interpreter.pipeline):
        p.process(msg)
        msg_dic = msg.as_dict()
        print("%2d. %-30s " %(i+1, type(p).__name__), end=" ")
        if not was_takens and 'text_tokens' in msg_dic:            # выводим токены:
            print("tokens:|", end='')
            for t in msg_dic['text_tokens']:
                print(t.text, end="|")               
            was_takens = True
        if type(p).__name__.find("Featurizer") >=0:
            print("features: ", end="")
            for f in msg.features:                        
                if str(type(f.features)).find('sparse') >= 0:
                    print('s', end="")
                else:
                    print('d', end="")
                print(f.features.shape, end=" ")
        print()
        fout.write("pipeline[%d]: %s  %s\n" % (i, type(p).__name__, type(msg).__name__))
        fout.write("data: "+str(msg.data)+"\n")        
        fout.write("features: "+str(msg.features)+"\n")                 
        fout.write("------------------------------------\n")
    fout.close()
    print("\n")

def get_intents(out):
    """Добавить информацию о возможных намерениях в строку out"""
    if 'intent' in result and 'name' in result['intent'] and 'confidence' in result['intent']:
        out += "%s [%.3f], " % (result['intent']['name'], result['intent']['confidence'])
    else:
        print("Something wrong... (no intent, name or confidence)")
        return out
    
    # выводим уверенность в остальных намерениях:
    if 'intent_ranking' in result:
        for i, intent in enumerate(result['intent_ranking']):        
            if intent['name'] == result['intent']['name']:
                continue
            out += ("%s [%.3f], " % (intent['name'],intent['confidence']))                        
            if i == 2:
                break                                    
    return out

def get_entities(out):
    """Информация о сущностях"""
    for i, k in enumerate(result['entities']):
        if 'value' in k:
            out += "\n%s=%s" % (k['entity'],k['value'])
        if 'confidence_entity' in k:
            out += " [%.2f]" % ( k['confidence_entity'] )
        if 'role' in k:
            out += ", role=%s" % ( k['role'] )
        if 'confidence_role' in k:
            out += " [%.2f]" % ( k['confidence_role'] )
        if 'grope' in k:
            out += ", group=%s" % ( k['group'] )
        if 'confidence_group' in k:
            out += " [%.2f]" % ( k['confidence_group'] )            
        if 'extractor' in k:
            out += ", %s" % ( k['extractor'] )
        
        if i+1 < len(result['entities']):        
            out += "; "
    return out

################################################################################

interpreter = load_interpreter()
inter_all   = interpreter.interpreter 
regex_interpreter = RegexInterpreter()

print_success("NLU model loaded. Type a message and press enter to parse it.")

while True:
    print_success("next message:")
    message = input()  # .replace("?", " query ")   # .lower()                     
        
    print()
    #get_tokens(TEXT);                      #  выводим токены
    get_pipeline()
       
    if message.startswith(INTENT_MESSAGE_PREFIX):
        print("loop")
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(regex_interpreter.parse(message))
    else:
        result = inter_all.parse(message)
            
    #print(result, end="\n\n")                      
    print(result['text'])
    
    print( get_intents("") )
    print( get_entities("") )   

"""
{'text': 'зовите меня маша', 
'intent': {'id': 5846425762300176186, 'name': 'my_name', 'confidence': 0.9994565844535828}, 
'entities': [{'entity': '"name"', 'start': 12, 'end': 16, 'confidence_entity': 0.9980849027633667, 
              'value': 'маша', 'extractor': 'DIETClassifier'}], 
'intent_ranking': [{'id': 5846425762300176186, 'name': 'my_name', 'confidence': 0.9994565844535828}, 
{'id': -3183177557687976971, 'name': 'deny', 'confidence': 0.00016833457630127668}, 
{'id': 1916181083327477722, 'name': 'affirm', 'confidence': 9.996127482736483e-05}, 
{'id': 8690575099665227747, 'name': 'mood_unhappy', 'confidence': 9.281196253141388e-05}, 
{'id': -5610299248971807086, 'name': 'greet', 'confidence': 8.362063817912713e-05}, 
{'id': -923989697737468401, 'name': 'mood_great', 'confidence': 4.0231429011328146e-05}, 
{'id': 3160065119479831266, 'name': 'bot_challenge', 'confidence': 2.964206214528531e-05}, 
{'id': -4155911471353619244, 'name': 'goodbye', 'confidence': 2.8788572308258153e-05}], 
'response_selector': {'all_retrieval_intents': [], 
'default': {'response': {'id': None, 'responses': None, 'response_templates': None, 
'confidence': 0.0, 'intent_response_key': None, 
'utter_action': 'utter_None', 'template_name': 'utter_None'}, 'ranking': []}}}
"""