"""
                        Анализиатор качества обучения nlu-модели.
                                    
Запускается после обучения:  rasa train nlu
Если в корне есть файл !nlu_stat.txt,  то после проверки данных обучения, 
тестируются тестовые примеры. Кроме вывода на экран, результаты помещаются в файл !nlu_stat.out                                                             
"""
from rasa.shared.constants         import DEFAULT_MODELS_PATH
from rasa.cli.utils                import get_validated_path
from rasa.model                    import get_model, get_model_subdirectories
from rasa.nlu.model                import Interpreter
from rasa.shared.nlu.interpreter   import RegexInterpreter, INTENT_MESSAGE_PREFIX
from rasa.shared.utils.cli         import print_success

import rasa.nlu.run

import markdown                                # https://python-markdown.github.io/reference/
from   bs4 import BeautifulSoup                # https://www.crummy.com/software/BeautifulSoup/bs4/doc/
import os.path
###############################################################
# Загружаем nlu-модель
#
model = get_validated_path("", "model", DEFAULT_MODELS_PATH)
model_path = get_model(model)
_, nlu_model = get_model_subdirectories(model_path)
    
interpreter = Interpreter.load(nlu_model, None)
regex_interpreter = RegexInterpreter()                         

###############################################################
# файл с результатами:
fout = open("!nlu_stat.res", 'w', encoding='utf-8')  
# Вывод на экран и в файл
def out (st):
    print(st)
    fout.write(st+"\n")
    
###############################################################
# Прочитать все интенты в словарь intents из yml-фала fname
# Каждый элемент словаря с именем intent - это массив примеров
#
def read_intends(intents, fname):

    import yaml
    from yaml import Loader as YamlLoader, Dumper as YamlDumper
     

    if not os.path.exists(fname) :
       print("no file", fname)
       return False
       
    f    = open(fname, 'r', encoding='utf-8')      # читаем yml-файл
    txt  = f.read()
    data = yaml.load(txt, Loader=YamlLoader)

    for intent in data['nlu']:        
        if not 'intent' in intent:
             continue                              # lookup, synonym
            
        name = intent['intent'].strip()
        list = intent['examples'].split('- ')
        intents[name] = { "strs":[], "acc":0, "min":0, "aver":0 }
        strs =  intents[name]["strs"]
        for li in list:
            str = li.strip()
            if len(str) > 0:
                strs.append( {"str": str, "info":""})    # добавляем их в массив интента
        strs.sort(key = lambda x:x["str"])               # сортируем по примерам
        
        if len(strs) > 1 :
            prev = strs[0]["str"]
            for k in range(1,len(strs)) :            
                if strs[k]["str"] == prev:
                    out("!!! Duplicate example '%s' in intent '%s'" % (prev, intent))
                prev = strs[k]["str"]
    
    return True
    
###############################################################
# Получаем наиболее вероятные интенты
#
def get_intents(message):
    if message.startswith(INTENT_MESSAGE_PREFIX):
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(regex_interpreter.parse(message))
    else:
        result = interpreter.parse(message)
       
    info = ""
    printedIntents = 0
    best, conf = "", 0
    for intent in result['intent_ranking']:            
        if printedIntents == 0 :
            best = intent['name']
            conf = intent['confidence']
          
        info += (", %-15s (%.3f)" % (intent['name'],intent['confidence']))                        
        printedIntents += 1      
        if printedIntents == 3:                     # число выводимых интентов
            break
          
    return best, conf, info
###############################################################
#
def analize(intents):   
    for intent in intents :
       num_ok, num_err, min, aver, min_sent = 0,0, 100, 0, ""
       for s in intents[intent]["strs"]:
          best, conf, info = get_intents(s["str"])      
          if best == intent : 
             num_ok  += 1
          else: 
             num_err += 1 
          if min > conf : min, min_sent = conf, s["str"]
          aver += conf
          s["info"] = info
       intents[intent]["acc"]  = num_ok/(num_ok+num_err)   
       intents[intent]["min"]  = min
       intents[intent]["min_sent"]  = min_sent
       intents[intent]["aver"] = aver/(num_ok+num_err)

###############################################################
# Выводим информацию
#
def out_stat(intents):   
    for intent in intents :   
        out("\n%-15s\tacc:%.3f min:%.3f" % (intent, intents[intent]["acc"], intents[intent]["min"]))   
        for s in intents[intent]["strs"]:
            out("\t%-45s\t%s" % (s["str"], s["info"]))
            
    out("---------------------------------------------------------")
    tot_aver = 0
    for intent in intents :        
        out("%-18s\tacc:%.3f  aver:%.3f   min:%.3f :  %-s" % (intent, intents[intent]["acc"], intents[intent]["aver"],  intents[intent]["min"], intents[intent]["min_sent"]))
        tot_aver += intents[intent]["aver"]
    out("---------------------------------------------------------")
    out("tot_aver: %.3f" % (tot_aver/len(intents)))
    out("---------------------------------------------------------")

###############################################################
#                              MAIN     
###############################################################
intents = {}                                         # список интентов   
if not read_intends(intents, 'data/nlu.yml') :       # получаем обученные интенты    
    exit()
out("\n========================= nlu.yml ===========================")

analize(intents)
out_stat(intents)

tests   = {}
if not read_intends(tests,   '!nlu_stat.txt'):  # получаем тестовые интенты 
    fout.close()
    exit()
    
out("\n========================= tests nlu_tests.txt ===========================")
analize(tests)
out_stat(tests)                                      # выводим информацию

fout.close()